<?php

namespace App\Http\Controllers\Screenwriter\ScreenwriterScripts;

use App\Events\UploadFiles;
use App\Http\Controllers\Controller;
use App\Models\Checklist;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use App\Models\Script;
use App\Services\CloudStorage\CloudStorageServices;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Storage;

use function PHPUnit\Framework\isEmpty;

class ScreenwriterScriptsController extends Controller
{
    use CloudStorageServices;
    public function __construct()
    {
        $this->middleware(['auth']);
        $this->middleware(['role:screenwriter']);
        $this->middleware(['dowalla.auth:Required']);
    }

    public function index()
    {

        $current_user = Auth::user();

        // Full Name of User
        $fullname = $current_user->fullname;

        // User Scripts
        $scripts = Script::where('user_id', $current_user->id);

        if ($scripts) {

            $scripts = $scripts->where('publish_status', 'draft')
                ->orWhere('publish_status', 'submitted')
                ->orWhere('publish_status', 'in_review')
                ->orWhere('publish_status', 'denied')
                ->orWhere('publish_status', 'accepted')
                ->orderBy('updated_at', 'DESC')->get();
        } else {
            $scripts = false;
        }

        return view('screenwriter.screenwriter_scripts.index', [
            'fullname' => $fullname,
            'scripts' => $scripts,
        ]);
    }

    public function upload_script(Request $request)
    {

        // Calling an event to fire pubsub

        $this->validate($request, [
            'genre' => 'required',
            'pages' => 'required',
            'mpaa_rating' => 'required',
            'budget' => 'required',
            'timeline' => 'required',
            'where' => 'required',
            'title' => 'required',
            'synopsis' => 'required',
            'tagline' => 'required',
            'logline' => 'required',
            'treatment' => 'required|mimes:pdf|max:1024',
            'script' => 'required|mimes:pdf|max:1024'
        ]);

        if ($request->upload == 'draft_upload') {
            $status = 'draft';
            $submissions = 0;
        } else if ($request->upload == 'review_upload') {
            $status = 'submitted';
            $submissions = 1;
        }
        //Adjusting the File Name
        $treatment_name = pathinfo($request->file('treatment')->getClientOriginalName(), PATHINFO_FILENAME);
        $treatment_unique_name = (time() * 13) . $treatment_name . '.pdf';

        $script_name = pathinfo($request->file('treatment')->getClientOriginalName(), PATHINFO_FILENAME);
        $script_unique_name = (time() * 13) . $script_name . '.pdf';

        $script = Script::create([
            'genre' => $request->genre,
            'pages' => $request->pages,
            'content_ratings' => $request->mpaa_rating,
            'budget' => $request->budget,
            'timeline' => $request->timeline,
            'where' => $request->where,
            'title' => $request->title,
            'synopsis' => $request->synopsis,
            'tagline' => $request->tagline,
            'logline' => $request->logline,
            'user_id' => Auth::user()->id,
            'publish_status' => $status,
            'treatment' => $treatment_unique_name,
            'script' => $script_unique_name,
            'submissions' => $submissions,
        ]);

        if ($request->upload == 'draft_upload') {
            Checklist::where('user_id', Auth::user()->id)
                ->update(['first_script' => true]);
        }

        // Store into Application Storage
        $request->file('treatment')->storeAs('', $treatment_unique_name, 'local');
        $request->file('script')->storeAs('', $script_unique_name, 'local');

        // Initilaze Cloud Storage Client and choosing Bucket
        $client = self::CloudStorageConfiguration();
        $bucket = $client->bucket(env('GOOGLE_CLOUD_STORAGE_BUCKET'));

        // For Treatment FIle Upload
        $bucket->upload(
            fopen(storage_path('app') . '/' . $treatment_unique_name, 'r'),
            [
                'name' => 'treatments/' . $treatment_unique_name . '',
            ]
        );

        // For Script FIle Upload
        $bucket->upload(
            fopen(storage_path('app') . '/' . $script_unique_name, 'r'),
            [
                'name' => 'scripts/' . $script_unique_name . '',
            ]
        );

        // Deleting the File from application storage
        Storage::disk('local')->delete($treatment_unique_name, $script_unique_name);
    }
}
