<?php

namespace App\Models;

use App\Jobs\CheckoutExpire;
use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\SoftDeletes;
use Illuminate\Support\Carbon;
// use Laravel\Cashier\Cashier;

class Script extends Model
{
    use HasFactory, SoftDeletes;

    protected $casts = [
        'title' => 'string',
        'synopsis' => 'string',
        'logline' => 'string',
        'tagline' => 'string',
        'user_id' => 'string',
        'treatment' => 'string',
        'script' => 'string',
        'optioned_until' => 'datetime',
        'published_at' => 'datetime',
        'application_id' => 'string',
        'application_type' => 'string',
        'pages' => 'string',
        'where' => 'string',
        'budget' => 'string',
        'timeline' => 'string',
        'genre' => 'string',
        'content_ratings' => 'string',
        'publish_status' => 'string',
    ];

    protected $fillable = [
        'title',
        'synopsis',
        'logline',
        'tagline',
        'pages',
        'budget',
        'where',
        'genre',
        'timeline',
        'content_ratings',
        'user_id',
        'publish_status',
        'treatment',
        'script',
        'submissions'
    ];

    protected $hidden = [
        'treatment',
        'screenplay',
        'deleted_at',
    ];

    protected $appends =['status'];

    public function author()
    {
        return $this->belongsTo(User::class, 'user_id');
    }

    /**
     * Get the rating associated with the script.
     */
    public function scriptRatings()
    {
        return $this->hasMany(ScriptRating::class)->latest();
    }

    public function budget()
    {
        return $this->belongsTo(BudgetOption::class);
    }

    public function timeline()
    {
        return $this->belongsTo(Timeline::class);
    }

    public function contentRating()
    {
        return $this->belongsTo(ContentRating::class);
    }

    public function genres()
    {
        return $this->belongsToMany(Genre::class);
    }

    public function orders()
    {
        return $this->hasMany(ScriptOrder::class);
    }

    public function esignatures()
    {
        return $this->hasMany(SignatureRequest::class, 'script_id', 'id');
    }

    public function application()
    {
        return $this->morphTo();
    }

    public function getTotalScriptRatingsAttribute()
    {
        $total_ratings = 0;

        if (!$this->scriptRatings->count()) {
            return 0;
        }

        $this->scriptRatings->each(function ($item) use (&$total_ratings) {
            $total_ratings += $item->totalRating();
        });

        return $total_ratings / $this->scriptRatings->count();
    }

    public function checkout(
        User   $user,
        string $success_url,
        string $cancel_url,
        string $checkout,
        string $period = '6_months'
    ) {
        $seller = $this->author->seller_account['id'];
        $seller_prices = collect($this->author->seller_prices);

        switch ($checkout) {
            case 'screenplay':
                $orderData = ['checkout' => 'screenplay', 'period' => $period];
                if ($user->account_type == User::ACCOUNT_TYPE_GUILD_PRODUCER) {
                    $priceId = $period == '6_months' ?
                        $seller_prices->where('nickname', 'screenplay_1')->first()['id'] :
                        $seller_prices->where('nickname', 'screenplay_2')->first()['id'];
                } else {
                    $priceId = $period == '6_months' ?
                        $seller_prices->where('nickname', 'screenplay_3')->first()['id'] :
                        $seller_prices->where('nickname', 'screenplay_4')->first()['id'];
                }
                break;
            case 'extension':
                $orderData = ['checkout' => 'extension', 'period' => '6_months'];
                $priceId = $user->account_type == User::ACCOUNT_TYPE_GUILD_PRODUCER ?
                    $seller_prices->where('nickname', 'extension_1')->first()['id'] :
                    $seller_prices->where('nickname', 'extension_2')->first()['id'];
                break;
            default:
                $orderData = ['checkout' => 'treatment'];
                $priceId = $seller_prices->where('nickname', 'treatment')->first()['id'];
                break;
        }

        $scriptOrder = $this->orders()->create([
            'user_id' => $user->id,
            ...$orderData
        ]);

        $now = Carbon::now();
        $stripeSessionExpires = $now->addHour()->timestamp;
        $waSessionExpires = $now->addMinutes(5);

        // $checkoutSession = Cashier::stripe()->checkout->sessions->create([
        //     'success_url' => $success_url,
        //     'cancel_url' => $cancel_url,
        //     'line_items' => [['price' => $priceId, 'quantity' => 1]],
        //     'mode' => 'payment',
        //     'metadata' => ['order_id' => $scriptOrder->id],
        //     'expires_at' => $stripeSessionExpires
        // ], ['stripe_account' => $seller]);

        // $scriptOrder->stripe_id = $checkoutSession->id;
        // $scriptOrder->checkout_url = $checkoutSession->url;
        // $scriptOrder->checkout_until = $waSessionExpires;
        // $scriptOrder->checkout_amount = $checkoutSession->amount_total;
        // $scriptOrder->save();

        CheckoutExpire::dispatch($checkoutSession)->delay($waSessionExpires);

        return $scriptOrder->fresh();
    }

    public function status()
    {
        if ($this->optioned_until && $this->optioned_until < Carbon::now()) {
            return 'Optioned';
        } elseif ($this->published_at) {
            return 'Published';
        } elseif (!$this->submitted_at && !$this->optioned_until && !$this->published_at && !$this->rejected_at && !$this->accepted_at) {
            return 'Draft';
        } elseif ($this->rejected_at && $this->submitted_at) {
            return 'Rejected';
        } elseif ($this->accepted_at && $this->submitted_at) {
            return 'Accepted';
        } elseif ($this->submitted_at && !$this->rejected_at && !$this->accepted_at) {
            return 'Submitted';
        } elseif ($this->submitted_at && !$this->optioned_until && !$this->published_at && !$this->rejected_at && !$this->accepted_at) {
            return 'In Review';
        }
    }

    public function isPurchased()
    {
        return $this->orders()->whereNotNull('checkout_completed')->first() ?? false;
    }

    public function isOptioned()
    {
        return $this->optioned_until < Carbon::now();
    }

    public function getScreenplayAvailableAttribute()
    {
        if ($this->isPurchased()) {
            return false;
        }

        if ($this->isOptioned()) {
            return false;
        }

        return true;
    }

    public function getTreatmentAvailableAttribute()
    {
        if ($this->isPurchased()) {
            return false;
        }

        if ($this->isOptioned()) {
            return false;
        }

        return true;
    }

    public static function updatePublishStatus(Script $script,$status){
        $script->publish_status = $status;
        $script->signature_requested_at = now();
        $script->save();
    }

    public static function publish(Script $script)
    {
        if($script->accepted_at){
            $script->pulished_at = now();
            if($script->save())
            {
                return true;
            }
        }else{
            return 'Not Accepted';
        }
    }

    public function getStatusAttribute()
    {
        if ($this->optioned_until && $this->optioned_until < Carbon::now()) {
            return 'Optioned';
        } elseif ($this->published_at) {
            return 'Published';
        } elseif (!$this->submitted_at && !$this->optioned_until && !$this->published_at && !$this->rejected_at && !$this->accepted_at) {
            return 'Draft';
        } elseif ($this->rejected_at && $this->submitted_at) {
            return 'Rejected';
        } elseif ($this->accepted_at && $this->submitted_at) {
            return 'Accepted';
        } elseif ($this->submitted_at && !$this->rejected_at && !$this->accepted_at) {
            return 'Submitted';
        } elseif ($this->submitted_at && !$this->optioned_until && !$this->published_at && !$this->rejected_at && !$this->accepted_at) {
            return 'In Review';
        }
    }
}
